# Copyright 2025 Daytona Platforms Inc.
# SPDX-License-Identifier: AGPL-3.0

from typing import Protocol


class SandboxCodeToolbox(Protocol):
    def get_default_image(self) -> str:
        ...

    def get_code_run_command(self, code: str) -> str:
        ...

    def get_code_run_args(self) -> list[str]:
        ...

    # ... other protocol methods



# Copyright 2025 Daytona Platforms Inc.
# SPDX-License-Identifier: AGPL-3.0

from daytona_api_client import SandboxVolume as ApiVolumeMount
from daytona_api_client import VolumeDto
from daytona_api_client_async import SandboxVolume as AsyncApiVolumeMount


class VolumeMount(ApiVolumeMount, AsyncApiVolumeMount):
    ...


class Volume(VolumeDto):
    """Represents a Daytona Volume which is a shared storage volume for Sandboxes.

    Attributes:
        id (StrictStr): Unique identifier for the Volume.
        name (StrictStr): Name of the Volume.
        organization_id (StrictStr): Organization ID of the Volume.
        state (StrictStr): State of the Volume.
        created_at (StrictStr): Date and time when the Volume was created.
        updated_at (StrictStr): Date and time when the Volume was last updated.
        last_used_at (StrictStr): Date and time when the Volume was last used.
    """

    @classmethod
    def from_dto(cls, dto: VolumeDto) -> "Volume":
        return cls(**dto.__dict__)



# Copyright 2025 Daytona Platforms Inc.
# SPDX-License-Identifier: AGPL-3.0

import warnings
from dataclasses import dataclass
from typing import Dict, List, Optional

from daytona_api_client import ExecuteResponse as ClientExecuteResponse
from daytona_api_client import SessionExecuteRequest as ApiSessionExecuteRequest
from daytona_api_client_async import SessionExecuteRequest as AsyncApiSessionExecuteRequest
from pydantic import ConfigDict, model_validator

from .charts import Chart


@dataclass
class CodeRunParams:
    """Parameters for code execution.

    Attributes:
        argv (Optional[List[str]]): Command line arguments
        env (Optional[Dict[str, str]]): Environment variables
    """

    argv: Optional[List[str]] = None
    env: Optional[Dict[str, str]] = None


class SessionExecuteRequest(ApiSessionExecuteRequest, AsyncApiSessionExecuteRequest):
    """Contains the request for executing a command in a session.

    Attributes:
        command (str): The command to execute.
        run_async (Optional[bool]): Whether to execute the command asynchronously.
        var_async (Optional[bool]): Deprecated. Use `run_async` instead.
    """

    @model_validator(mode="before")
    @classmethod
    def __handle_deprecated_var_async(cls, values):  # pylint: disable=unused-private-member
        if "var_async" in values and values.get("var_async"):
            warnings.warn(
                "'var_async' is deprecated and will be removed in a future version. Use 'run_async' instead.",
                DeprecationWarning,
                stacklevel=3,
            )
            if "run_async" not in values or not values["run_async"]:
                values["run_async"] = values.pop("var_async")
        return values


class ExecutionArtifacts:
    """Artifacts from the command execution.

    Attributes:
        stdout (str): Standard output from the command, same as `result` in `ExecuteResponse`
        charts (Optional[List[Chart]]): List of chart metadata from matplotlib
    """

    stdout: str
    charts: Optional[List[Chart]] = None

    def __init__(self, stdout: str = "", charts: Optional[List[Chart]] = None):
        self.stdout = stdout
        self.charts = charts


class ExecuteResponse(ClientExecuteResponse):
    """Response from the command execution.

    Attributes:
        exit_code (int): The exit code from the command execution
        result (str): The output from the command execution
        artifacts (Optional[ExecutionArtifacts]): Artifacts from the command execution
    """

    artifacts: Optional[ExecutionArtifacts] = None

    # TODO: Remove model_config once everything is migrated to pydantic # pylint: disable=fixme
    model_config = ConfigDict(arbitrary_types_allowed=True)

    # pylint: disable=super-init-not-called
    def __init__(
        self,
        exit_code: int,
        result: str,
        artifacts: Optional[ExecutionArtifacts] = None,
        additional_properties: Dict = None,
    ):
        self.exit_code = exit_code
        self.result = result
        self.additional_properties = additional_properties or {}
        self.artifacts = artifacts



# Copyright 2025 Daytona Platforms Inc.
# SPDX-License-Identifier: AGPL-3.0

from daytona_api_client import SandboxState, SessionExecuteResponse

from ._async.daytona import AsyncDaytona
from ._async.sandbox import AsyncSandbox
from ._sync.daytona import Daytona
from ._sync.sandbox import Sandbox

# Create deprecated aliases with proper warnings
from ._utils.deprecation import deprecated_alias
from .common.charts import (
    BarChart,
    BoxAndWhiskerChart,
    Chart,
    ChartType,
    CompositeChart,
    LineChart,
    PieChart,
    ScatterChart,
)
from .common.daytona import (
    CodeLanguage,
    CreateSandboxBaseParams,
    CreateSandboxFromImageParams,
    CreateSandboxFromSnapshotParams,
    DaytonaConfig,
)
from .common.errors import DaytonaError
from .common.filesystem import FileUpload
from .common.image import Image
from .common.lsp_server import LspLanguageId
from .common.process import CodeRunParams, SessionExecuteRequest
from .common.sandbox import Resources
from .common.snapshot import CreateSnapshotParams
from .common.volume import VolumeMount

__all__ = [
    "Daytona",
    "DaytonaConfig",
    "CodeLanguage",
    "SessionExecuteRequest",
    "SessionExecuteResponse",
    "DaytonaError",
    "LspLanguageId",
    "CodeRunParams",
    "Sandbox",
    "Resources",
    "SandboxState",
    "ChartType",
    "Chart",
    "LineChart",
    "ScatterChart",
    "BarChart",
    "PieChart",
    "BoxAndWhiskerChart",
    "CompositeChart",
    "FileUpload",
    "VolumeMount",
    "AsyncDaytona",
    "AsyncSandbox",
    "Image",
    "CreateSandboxBaseParams",
    "CreateSandboxFromImageParams",
    "CreateSandboxFromSnapshotParams",
    "CreateSnapshotParams",
]



# Copyright 2025 Daytona Platforms Inc.
# SPDX-License-Identifier: AGPL-3.0

import asyncio
import hashlib
import os
import tarfile
import threading

import aioboto3

from .._utils.docs_ignore import docs_ignore


class AsyncObjectStorage:
    """AsyncObjectStorage class for interacting with object storage services.

    Attributes:
        endpoint_url (str): The endpoint URL for the object storage service.
        aws_access_key_id (str): The access key ID for the object storage service.
        aws_secret_access_key (str): The secret access key for the object storage service.
        aws_session_token (str): The session token for the object storage service. Used for temporary credentials.
        bucket_name (str): The name of the bucket to use.
    """

    def __init__(
        self,
        endpoint_url,
        aws_access_key_id,
        aws_secret_access_key,
        aws_session_token,
        bucket_name="daytona-volume-builds",
    ):
        self.bucket_name = bucket_name
        self.endpoint_url = endpoint_url
        self.s3_client = aioboto3.Session().client(
            service_name="s3",
            endpoint_url=self.endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
        )
        # unasync: delete start
        self._client_ctx = None

    async def open_client(self):
        try:
            self._client_ctx = self.s3_client
            # pylint: disable=unnecessary-dunder-call
            self.s3_client = await self._client_ctx.__aenter__()
        except Exception as e:
            raise Exception(f"Error opening S3 client: {e}") from e

    async def close_client(self):
        try:
            await self._client_ctx.__aexit__(None, None, None)
            self.s3_client = None
            self._client_ctx = None
        except Exception as e:
            raise Exception(f"Error closing S3 client: {e}") from e

    # unasync: delete end

    async def upload(self, path, organization_id, archive_base_path=None) -> str:
        """Uploads a file to the object storage service.

        Args:
            path (str): The path to the file to upload.
            organization_id (str): The organization ID to use.
            archive_base_path (str): The base path to use for the archive.
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"Path does not exist: {path}")

        # Compute hash for the path
        path_hash = self._compute_hash_for_path_md5(path, archive_base_path)

        # Define the S3 prefix
        prefix = f"{organization_id}/{path_hash}/"
        s3_key = f"{prefix}context.tar"

        # Check if it already exists in S3
        if await self._folder_exists_in_s3(prefix):
            return path_hash

        # Upload to S3
        await self._upload_as_tar(s3_key, path, archive_base_path)

        return path_hash

    @staticmethod
    @docs_ignore
    def compute_archive_base_path(path_str) -> str:
        """Compute the base path for an archive. Returns normalized path without the root
        (drive letter or leading slash).

        Args:
            path_str (str): The path to compute the base path for.

        Returns:
            str: The base path for the given path.
        """
        path_str = os.path.normpath(path_str)
        # Remove drive letter for Windows paths (e.g., C:)
        _, path_without_drive = os.path.splitdrive(path_str)
        # Remove leading separators (both / and \)
        return path_without_drive.lstrip("/").lstrip("\\")

    def _compute_hash_for_path_md5(self, path_str, archive_base_path=None):
        """Computes the MD5 hash for a given path.

        Args:
            path_str (str): The path to compute the hash for.
            archive_base_path (str): The base path to use for the archive.

        Returns:
            str: The MD5 hash for the given path.
        """
        md5_hasher = hashlib.md5()
        abs_path_str = os.path.abspath(path_str)

        if archive_base_path is None:
            archive_base_path = self.compute_archive_base_path(path_str)
        md5_hasher.update(archive_base_path.encode("utf-8"))

        if os.path.isfile(abs_path_str):
            with open(abs_path_str, "rb") as f:
                for chunk in iter(lambda: f.read(8192), b""):
                    md5_hasher.update(chunk)
        else:
            for root, dirs, files in os.walk(abs_path_str):
                if not dirs and not files:
                    rel_dir = os.path.relpath(root, path_str)
                    md5_hasher.update(rel_dir.encode("utf-8"))
                for filename in files:
                    file_path = os.path.join(root, filename)
                    rel_path = os.path.relpath(file_path, abs_path_str)

                    # Incorporate the relative path
                    md5_hasher.update(rel_path.encode("utf-8"))

                    # Incorporate file contents
                    with open(file_path, "rb") as f:
                        for chunk in iter(lambda: f.read(8192), b""):
                            md5_hasher.update(chunk)

        return md5_hasher.hexdigest()

    async def _folder_exists_in_s3(self, prefix):
        """Checks if a folder exists in S3.

        Args:
            prefix (str): The prefix to check.

        Returns:
            bool: True if the folder exists, False otherwise.
        """
        resp = await self.s3_client.list_objects_v2(Bucket=self.bucket_name, Prefix=prefix)
        return "Contents" in resp

    async def _upload_as_tar(self, s3_key, source_path, archive_base_path=None):
        """Uploads a file to the object storage service as a tar.

        Args:
            s3_key (str): The key to upload the file to.
            source_path (str): The path to the file to upload.
            archive_base_path (str): The base path to use for the archive.
        """
        source_path = os.path.normpath(source_path)

        if archive_base_path is None:
            archive_base_path = self.compute_archive_base_path(source_path)

        read_fd, write_fd = os.pipe()
        read_file = os.fdopen(read_fd, "rb")
        write_file = os.fdopen(write_fd, "wb")

        def tar_worker():
            with tarfile.open(fileobj=write_file, mode="w|") as tar:
                tar.add(source_path, arcname=archive_base_path)
            write_file.close()

        thread = threading.Thread(target=tar_worker, daemon=True)
        thread.start()

        await self.s3_client.upload_fileobj(Fileobj=read_file, Bucket=self.bucket_name, Key=s3_key)

        read_file.close()
        await asyncio.to_thread(thread.join)



# Copyright 2025 Daytona Platforms Inc.
# SPDX-License-Identifier: AGPL-3.0

import concurrent.futures
import functools
from typing import Any, Callable, Optional, ParamSpec, TypeVar

from .._utils.errors import DaytonaError

P = ParamSpec("P")
T = TypeVar("T")


def with_timeout(
    error_message: Optional[Callable[[Any, float], str]] = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator to add a timeout mechanism with an optional custom error message.

    Args:
        error_message (Optional[Callable[[Any, float], str]]): A callable that accepts `self` and `timeout`,
                                                               and returns a string error message.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            # Get function argument names
            arg_names = func.__code__.co_varnames[: func.__code__.co_argcount]
            arg_dict = dict(zip(arg_names, args))

            # Extract self if method is bound
            self_instance = args[0] if args else None

            # Check for 'timeout' in kwargs first, then in positional arguments
            timeout = kwargs.get("timeout", arg_dict.get("timeout", None))

            if timeout is None or timeout == 0:
                # If timeout is None or 0, run the function normally
                return func(*args, **kwargs)

            if timeout < 0:
                raise DaytonaError("Timeout must be a non-negative number or None.")

            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(func, *args, **kwargs)
                try:
                    return future.result(timeout=timeout)
                except concurrent.futures.TimeoutError:
                    # Use custom error message if provided, otherwise default
                    msg = (
                        error_message(self_instance, timeout)
                        if error_message
                        else f"Function '{func.__name__}' exceeded timeout of {timeout} seconds."
                    )
                    raise TimeoutError(msg)  # pylint: disable=raise-missing-from

        return wrapper

    return decorator



# Copyright 2025 Daytona Platforms Inc.
# SPDX-License-Identifier: AGPL-3.0

import os
from pathlib import PurePosixPath
from typing import Optional


def prefix_relative_path(prefix: str, path: Optional[str] = None) -> str:
    result = prefix

    if path:
        path = path.strip()
        if path == "~":
            result = prefix
        elif path.startswith("~/"):
            result = os.path.join(prefix, path[2:])
        elif PurePosixPath(path).is_absolute():
            result = path
        else:
            result = os.path.join(prefix, path)

    return result



# Copyright 2025 Daytona Platforms Inc.
# SPDX-License-Identifier: AGPL-3.0

from enum import Enum
from typing import Optional


def to_enum(enum_class: type, value: str) -> Optional[Enum]:
    """Convert a string to an enum.

    Args:
        enum_class (type): The enum class to convert to.
        value (str): The value to convert to an enum.

    Returns:
        The enum value, or None if the value is not a valid enum.
    """
    if isinstance(value, enum_class):
        return value
    str_value = str(value)
    if str_value in enum_class._value2member_map_:  # pylint: disable=protected-access
        return enum_class(str_value)
    return None



# Copyright 2025 Daytona Platforms Inc.
# SPDX-License-Identifier: AGPL-3.0

import warnings
from enum import Enum, EnumMeta
from typing import Any, Callable, ParamSpec, Type, TypeVar, cast

E = TypeVar("E", bound=Enum)
P = ParamSpec("P")
T = TypeVar("T")


def deprecated_alias(old_name: str, new_name: str) -> Callable[[Type[T]], Type[T]]:
    """Decorator to mark a class or enum as deprecated with an alias.

    Args:
        old_name (str): The deprecated name/alias
        new_name (str): The new name that should be used instead

    Returns:
        A wrapped class that issues deprecation warnings when used
    """

    def decorator(cls: Type[T]) -> Type[T]:
        # Create warning message once
        warning_message = (
            f"`{old_name}` is deprecated. Please use `{new_name}` instead. "
            + "This will be removed in a future version."
        )

        if isinstance(cls, type) and issubclass(cls, Enum):

            class DeprecatedEnumMeta(EnumMeta):  # pylint: disable=unused-variable
                def __getattribute__(cls, name: str) -> Any:
                    if not name.startswith("_"):
                        warnings.warn(warning_message, DeprecationWarning, stacklevel=2)
                    return super().__getattribute__(name)

            # Create the deprecated enum class with optimized creation
            class DeprecatedEnum(Enum, metaclass=DeprecatedEnumMeta):
                def __new__(cls, value: Any) -> "DeprecatedEnum":
                    obj = object.__new__(cls)
                    obj._value_ = value
                    return obj

                def __eq__(self, other: Any) -> bool:
                    return self.value == getattr(other, "value", other)

            # Add enum members and copy metadata in one pass
            for item in cls:
                setattr(DeprecatedEnum, item.name, item.value)

            # Copy metadata attributes directly
            for attr in ("__module__", "__qualname__", "__name__", "__doc__"):
                setattr(
                    DeprecatedEnum,
                    attr,
                    getattr(cls, attr) if attr != "__name__" else old_name,
                )

            return cast(Type[T], DeprecatedEnum)

        # For non-enum classes, create a wrapper class that preserves type hints
        class WrappedClass(cls):  # type: ignore
            def __new__(cls, *args: P.args, **kwargs: P.kwargs) -> T:
                warnings.warn(warning_message, DeprecationWarning, stacklevel=2)
                return super().__new__(cls)  # pylint: disable=no-value-for-parameter

            def __init__(self, *args: P.args, **kwargs: P.kwargs) -> None:
                warnings.warn(warning_message, DeprecationWarning, stacklevel=2)
                super().__init__(*args, **kwargs)

        # Copy class attributes and metadata
        WrappedClass.__name__ = old_name
        WrappedClass.__qualname__ = cls.__qualname__
        WrappedClass.__module__ = cls.__module__
        WrappedClass.__doc__ = cls.__doc__

        # Copy annotations if they exist
        if hasattr(cls, "__annotations__"):
            WrappedClass.__annotations__ = dict(cls.__annotations__)

        # Copy any additional attributes from the original class
        for attr, value in cls.__dict__.items():
            if not attr.startswith("__"):
                setattr(WrappedClass, attr, value)

        return cast(Type[T], WrappedClass)

    return decorator
